import click
from pyETA.application import main as main_application
from pyETA.components.track import main as main_track
from pyETA.components.window import main as main_window
from pyETA.components.validate import main as main_validate

@click.group()
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode. (Creates debug.log file(s))",
)
@click.version_option(package_name="pyETA-toolbox")
def main(debug):
    if debug:
        from pyETA import LOGGER
        import logging
        import logging.handlers
        LOG_FORMAT = '%(asctime)s :: %(name)s:%(filename)s:%(funcName)s:%(lineno)d :: %(levelname)s :: %(message)s'
        file_handler = logging.handlers.RotatingFileHandler(
            filename="debug.log",
            mode='w',
            maxBytes=1000000, # 1MB
            encoding='utf-8', backupCount=2)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        LOGGER.addHandler(file_handler)


main.add_command(main_application, name="application")
main.add_command(main_track, name="track")
main.add_command(main_window, name="window")
main.add_command(main_validate, name="validate")

if __name__ == "__main__":
    main()


import argparse
import sys
import datetime
import os
import json
import time
import click
import threading
import pylsl

from typing import Optional
from mne_lsl import lsl
from collections import namedtuple
from dataclasses import dataclass
from pyETA import __datapath__, LOGGER
from pyETA.components import mock
from pyETA.components import utils
try:
    import tobii_research as tr
except ModuleNotFoundError:
    LOGGER.error("Without tobii_research library, Tobii eye-tracker won't work.")
import numpy as np

@dataclass
class FixationTuple:
    is_fixated: bool = False
    velocity: float = 0.0
    x: float = 0.0
    y: float = 0.0
    filtered_x: float = 0.0
    filtered_y: float = 0.0
    timestamp: float = 0.0
    elapsed_time: float = 0.0
    duration: float = 0.0

class Tracker:
    def __init__(
            self,
            data_rate=600,
            use_mock=False,
            fixation=False,
            velocity_threshold=0.5,
            accept_screen_nans=False,
            verbose=False,
            push_stream=False,
            save_data=False,
            screen_index=0,
            **kwargs
        ):
        self.screen_index = screen_index
        self.screen_width, self.screen_height = utils.get_current_screen_size(self.screen_index)
        self.data_rate = data_rate
        self.accept_screen_nans = accept_screen_nans
        self.verbose = verbose
        self.save_data = save_data
        self.use_mock = use_mock
        self.fixation = fixation
        self.velocity_threshold = velocity_threshold
        self.push_stream = push_stream
        self.gaze_data = []
        self.gaze_id = None
        self.lsl_gaze_outlet = None
        self._break_flag = False
        self.id = None
        
        try:
            if self.use_mock:
                LOGGER.info("Using a mock service.")
                self.gaze_id = mock.EYETRACKER_GAZE_DATA
                
                self.eyetracker = mock.MockEyeTracker(data_rate=self.data_rate, verbose=self.verbose)
                self.eyetracker.start()
            else:
                LOGGER.info("Using tobii to find eyetrackers.")
                self.gaze_id = tr.EYETRACKER_GAZE_DATA
                self.eyetrackers = tr.find_all_eyetrackers()
                self.eyetracker = self.eyetrackers[0] if self.eyetrackers else None
            if self.eyetracker:
                LOGGER.info(f"Screen Resolution: {self.screen_width}x{self.screen_height}")
                LOGGER.info(f"Address: {self.eyetracker.address}")
                LOGGER.info(f"Model: {self.eyetracker.model}")
                LOGGER.info(f"Name: {self.eyetracker.device_name}")
                LOGGER.info(f"Serial number: {self.eyetracker.serial_number}")
            else:
                raise ValueError("No eye tracker device/Mock found.")
        except Exception as e:
            raise ValueError(f"Error initializing the eye tracker: {e}")
        if self.push_stream:
            stream_info = lsl.StreamInfo(
                name='tobii_gaze_fixation',
                stype='Gaze',
                n_channels=22,
                sfreq=self.data_rate,
                dtype='float64',
                source_id=self.eyetracker.serial_number)
            LOGGER.info(f"Stream name: {stream_info.name}")

            ch_info = [
                #left eye
                ('left_gaze_x', 'gaze', 'normalized'),
                ('left_gaze_y', 'gaze', 'normalized'),
                ('left_pupil_diameter', 'pupil', 'mm'),
                ('left_fixated', 'fixation', 'boolean'),
                ('left_velocity', 'velocity', 'px'),
                ('left_fixation_timestamp', 'timestamp', 's'),
                ('left_fixation_elapsed', 'duration', 's'),
                ('left_filtered_gaze_x', 'filtered_gaze', 'normalized'),
                ('left_filtered_gaze_y', 'filtered_gaze', 'normalized'),
                #right eye
                ('right_gaze_x', 'gaze', 'normalized'),
                ('right_gaze_y', 'gaze', 'normalized'),
                ('right_pupil_diameter', 'pupil', 'mm'),
                ('right_fixated', 'fixation', 'boolean'),
                ('right_velocity', 'velocity', 'px'),
                ('right_fixation_timestamp', 'timestamp', 's'),
                ('right_fixation_elapsed', 'duration', 's'),
                ('right_filtered_gaze_x', 'filtered_gaze', 'normalized'),
                ('right_filtered_gaze_y', 'filtered_gaze', 'normalized'),
                #screen data
                ('screen_width', 'screen', 'px'),
                ('screen_height', 'screen', 'px'),
                ('timestamp', 'timestamp', 's'),
                ('local_clock', 'timestamp', 's')
            ]
            
            ch_names, ch_types, ch_units = zip(*ch_info)
            stream_info.set_channel_names(ch_names)
            stream_info.set_channel_types(ch_types)
            stream_info.set_channel_units(ch_units)
            self.lsl_gaze_outlet = lsl.StreamOutlet(stream_info)
            LOGGER.info(f"LSL Stream Info: {self.lsl_gaze_outlet.get_sinfo()}")
        self.__fixation_left = FixationTuple()
        self.__fixation_right = FixationTuple()
        if self.fixation:
            min_cutoff = 1.5
            beta = 1
            self.__one_euro_filter = self.create_filter(
                min_cutoff,
                beta,
                elements=["left_eye_x", "left_eye_y", "right_eye_x", "right_eye_y"])
        LOGGER.debug(f"Member Variables: {vars(self)}")
        print("Press Ctrl+C to stop tracking...")
    
    def create_filter(self, min_cutoff, beta, elements: list):
        Filter = namedtuple('Filter', elements)
        objects = {}
        for keyword in elements:
            filter_element = utils.OneEuroFilter(
                initial_time=utils.get_timestamp(),
                initial_value=0.0,
                min_cutoff=min_cutoff,
                beta=beta
            )
            LOGGER.debug(f"{keyword}:  {filter_element.previous_time}")
            objects[keyword] = filter_element
        return Filter(**objects)

    def _update_fixation_data(self, t, x, y, element):
        def calculate(previous_t, t, filtered_x, filtered_y):
            LOGGER.debug(f"x: {x}, y: {y}, filtered_x: {filtered_x}, filtered_y: {filtered_y}")
            distance = np.sqrt((filtered_x - x)**2 + (filtered_y - y)**2)
            elapsed_time = t - previous_t
            velocity = distance / elapsed_time
            is_fixated = False
            if velocity <= self.velocity_threshold:
                is_fixated = True
            LOGGER.debug(f"Distance: {distance}, Elapsed_Time: {elapsed_time}, Velocity: {velocity}")
            return is_fixated, velocity, elapsed_time
        if element == "left_eye":
            previous_t = min(self.__one_euro_filter.left_eye_x.previous_time, self.__one_euro_filter.left_eye_y.previous_time)
            
            # filter for x
            filtered_x = self.__one_euro_filter.left_eye_x(t, x)
            # filter for y
            filtered_y = self.__one_euro_filter.left_eye_y(t, y)
            LOGGER.debug(f"{element}: {previous_t}, {t}, ({x}, {filtered_x}),  ({y}, {filtered_y})")
            is_fixated, velocity, elapsed_time = calculate(previous_t, t, filtered_x, filtered_y)
            self.__fixation_left.x = x
            self.__fixation_left.y = y
            self.__fixation_left.filtered_x = filtered_x
            self.__fixation_left.filtered_y = filtered_y
            self.__fixation_left.velocity = velocity
            self.__fixation_left.is_fixated = is_fixated
            self.__fixation_left.timestamp = t if not is_fixated else self.__fixation_left.timestamp
            self.__fixation_left.elapsed_time = elapsed_time
            self.__fixation_left.duration = self.__fixation_left.duration + elapsed_time if is_fixated else 0
        elif element == "right_eye":
            previous_t = min(self.__one_euro_filter.right_eye_x.previous_time, self.__one_euro_filter.right_eye_y.previous_time)
            
            # filter for x
            filtered_x = self.__one_euro_filter.right_eye_x(t, x)
            # filter for y
            filtered_y = self.__one_euro_filter.right_eye_y(t, y)
            LOGGER.debug(f"{element}: {previous_t}, {t}, ({x}, {filtered_x}),  ({y}, {filtered_y})")
            is_fixated, velocity, elapsed_time = calculate(previous_t, t, filtered_x, filtered_y)
            self.__fixation_right.x = x
            self.__fixation_right.y = y
            self.__fixation_right.filtered_x = filtered_x
            self.__fixation_right.filtered_y = filtered_y
            self.__fixation_right.velocity = velocity
            self.__fixation_right.is_fixated = is_fixated
            self.__fixation_right.timestamp = t if not is_fixated else self.__fixation_right.timestamp
            self.__fixation_right.elapsed_time = elapsed_time
            self.__fixation_right.duration = self.__fixation_right.duration + elapsed_time if is_fixated else 0
        else:
            raise ValueError(f"Unknown element: {element}")

    def _collect_gaze_data(self, gaze_data):
        timestamp = utils.get_timestamp()
        if self.fixation:
            # left eye
            self._update_fixation_data(
                    t=timestamp,
                    x=gaze_data.get("left_gaze_point_on_display_area")[0],
                    y=gaze_data.get("left_gaze_point_on_display_area")[1],
                    element="left_eye"
                )
            
            # right eye
            self._update_fixation_data(
                    t=timestamp,
                    x=gaze_data.get("right_gaze_point_on_display_area")[0],
                    y=gaze_data.get("right_gaze_point_on_display_area")[1],
                    element="right_eye"
                )
        data = {
            "timestamp": timestamp,
            "device_time_stamp": gaze_data.get("device_time_stamp"),
            "left_eye": {
                "gaze_point": gaze_data.get("left_gaze_point_on_display_area"),
                "pupil_diameter": gaze_data.get("left_pupil_diameter"),
                "fixated": self.__fixation_left.is_fixated,
                "velocity": self.__fixation_left.velocity,
                "filtered_gaze_point": [self.__fixation_left.filtered_x, self.__fixation_left.filtered_y],
                "fixation_timestamp": self.__fixation_left.timestamp,
                "fixation_elapsed": self.__fixation_left.duration,
            },
            "right_eye": {
                "gaze_point": gaze_data.get("right_gaze_point_on_display_area"),
                "pupil_diameter": gaze_data.get("right_pupil_diameter"),
                "fixated": self.__fixation_right.is_fixated,
                "velocity": self.__fixation_right.velocity,
                "filtered_gaze_point": [self.__fixation_right.filtered_x, self.__fixation_right.filtered_y],
                "fixation_timestamp": self.__fixation_right.timestamp,
                "fixation_elapsed": self.__fixation_right.duration,
            }
        }
        
        stream_data = np.array([
                    data["left_eye"]["gaze_point"][0] if data["left_eye"]["gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["left_eye"]["gaze_point"][1] if data["left_eye"]["gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["left_eye"]["pupil_diameter"] if data["left_eye"]["pupil_diameter"] else np.nan if self.accept_screen_nans else 0,
                    data["left_eye"]["fixated"],
                    data["left_eye"]["velocity"],
                    data["left_eye"]["fixation_timestamp"],
                    data["left_eye"]["fixation_elapsed"],
                    data["left_eye"]["filtered_gaze_point"][0] if data["left_eye"]["filtered_gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["left_eye"]["filtered_gaze_point"][1] if data["left_eye"]["filtered_gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["right_eye"]["gaze_point"][0] if data["right_eye"]["gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["right_eye"]["gaze_point"][1] if data["right_eye"]["gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["right_eye"]["pupil_diameter"] if data["right_eye"]["pupil_diameter"] else np.nan if self.accept_screen_nans else 0,
                    data["right_eye"]["fixated"],
                    data["right_eye"]["velocity"],
                    data["right_eye"]["fixation_timestamp"],
                    data["right_eye"]["fixation_elapsed"],
                    data["right_eye"]["filtered_gaze_point"][0] if data["right_eye"]["filtered_gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    data["right_eye"]["filtered_gaze_point"][1] if data["right_eye"]["filtered_gaze_point"] else np.nan if self.accept_screen_nans else 0,
                    self.screen_width,
                    self.screen_height,
                    data["timestamp"],
                    lsl.local_clock()
                ], dtype=np.float64)
        if self.push_stream:
            self.lsl_gaze_outlet.push_sample(stream_data)
        if self.save_data: self.gaze_data.append(data)
        def multiply_tuples(t1, t2=(self.screen_width, self.screen_height)):
            return tuple(x*y for x,y in zip(t1, t2))
        if self.verbose:
            print(f'L: {multiply_tuples(data["left_eye"]["gaze_point"])} -> {multiply_tuples(data["left_eye"]["filtered_gaze_point"])}, R: {multiply_tuples(data["right_eye"]["gaze_point"])} -> {multiply_tuples(data["right_eye"]["filtered_gaze_point"])}')
            print(f'L: ({data.get("left_eye").get("fixated")}, {data.get("left_eye").get("velocity")}, {data.get("left_eye").get("fixation_elapsed")}), R: ({data.get("right_eye").get("fixated")}, {data.get("right_eye").get("velocity")}, {data.get("right_eye").get("fixation_elapsed")})')
            print(f'({self.screen_width} x {self.screen_height})')
            print('\n')

    def start_tracking(self, duration: Optional[float]=None):
        """Starts tracking continuously and saves the data to a file, if save_data flag is set to True during initialization."""
        end_time = datetime.datetime.now() + datetime.timedelta(seconds=duration) if duration is not None else None
        
        LOGGER.debug(f"Tracking for {duration} seconds...") if duration else LOGGER.debug("Tracking continuously...")
        if self.eyetracker:
            self.id = threading.get_native_id()
            try:
                LOGGER.info("Starting tracking...")
                self.eyetracker.subscribe_to(self.gaze_id, self._collect_gaze_data, as_dictionary=True)
                while True:
                    if end_time and datetime.datetime.now() >= end_time or self._break_flag:
                        self.stop_tracking()
                        LOGGER.info("Tracking stopped from loop!")
                        self._break_flag = False
                        break
                    time.sleep(1)
            except KeyboardInterrupt:
                self.stop_tracking()
            finally:
                if self.save_data:
                    if not os.path.exists(__datapath__):
                        os.makedirs(__datapath__)
                    file = os.path.join(__datapath__, f"gaze_data_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.json")
                    with open(file, "w") as f:
                        json.dump(
                            {
                                "screen_size": (self.screen_width, self.screen_height),
                                "data": self.gaze_data
                            }, f, indent=4)
                    LOGGER.info(f"Gaze Data saved: {file}!")
        else:
            LOGGER.debug("No eye tracker found!")
    
    def signal_break(self):
        self._break_flag = True

    def stop_tracking(self):
        self.id = None
        try:
            if self.eyetracker:
                LOGGER.info("Stopping tracking...")
                self.eyetracker.unsubscribe_from(self.gaze_id, self._collect_gaze_data)
                if self.use_mock:
                    self.eyetracker.stop()
                if self.push_stream:
                    del self.lsl_gaze_outlet
                    self.lsl_gaze_outlet = None
            else:
                LOGGER.info("No eye tracker found!")
        except Exception as e:
            LOGGER.debug(f"Error stopping tracking: {e}")
        return True


@click.command(name="track")
@click.option("--push_stream", is_flag=True, help="Push the data to LSL stream.")
@click.option("--data_rate", default=600, type=int, help="The rate of the data stream.")
@click.option("--use_mock", is_flag=True, help="Use this to start the mock service")
@click.option("--fixation", is_flag=True, help="Use this to add fixations duration to the data stream")
@click.option("--velocity", type=float, default=1.5, help="The velocity threshold for fixation")
@click.option("--accept_screen_nans", is_flag=True, help="Use this to avoid correcting for NaNs")
@click.option("--save_data", is_flag=True, help="Save the data to a file")
@click.option("--verbose", is_flag=True, help="Use this to display LOGGER.debug statements")
@click.option("--duration", type=float, help="The duration for which to track the data")
@click.option("--screen_index", type=int, default=0, help="The index of the screen to use")
def main(push_stream, data_rate, use_mock, fixation, velocity, accept_screen_nans, save_data, verbose, duration, screen_index):
    LOGGER.debug(f"Arguments: {locals()}")
    tracker = Tracker(
        data_rate=data_rate,
        use_mock=use_mock,
        fixation=fixation,
        velocity_threshold=velocity,
        accept_screen_nans=accept_screen_nans,
        verbose=verbose,
        push_stream=push_stream,
        save_data=save_data,
        screen_index=screen_index
    )
    if duration:
        tracker.start_tracking(duration=duration)
    else:
        tracker.start_tracking()


if __name__ == '__main__':
    main()



import sys
import math
import platform
import datetime
from typing import List, Optional
import PyQt6.QtWidgets as qtw
import PyQt6.QtGui as qtg
import numpy as np
import os
import glob
import os
from pyETA import __datapath__, LOGGER

def get_current_screen_size(screen_index=0):
    app = qtw.QApplication.instance()
    app_created = False

    if app:
        LOGGER.info("This framework application already exists. Using an instance!")
    else:
        app = qtw.QApplication(sys.argv)
        app_created = True
        LOGGER.info("Created a new application to fetch the geometry!")

    screens = app.screens()

    try:
        if screen_index < len(screens):
            screen = screens[screen_index]
            geometry = screen.geometry()
            return geometry.width(), geometry.height()
        else:
            raise ValueError(f"Invalid screen index. Only {len(screens)} screen(s) available. (starting index from 0)")
    except Exception as e:
        LOGGER.error(str(e))
        raise e
    finally:
        if app_created:
            app.quit()
            LOGGER.info("Quitting the application created to fetch the geometry.")
    return 0, 0

def get_system_info():
    node = platform.node()
    system = platform.system()
    machine = platform.machine()
    time_now = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    return f"{node}_{system}_{machine}_{time_now}"

def get_timestamp():
    return datetime.datetime.now().timestamp()


class OneEuroFilter:
    def __init__(
        self,
        initial_time: float,
        initial_value: float,
        initial_derivative: float = 0.0,
        min_cutoff: float = 1.0,
        beta: float = 0.0,
        derivative_cutoff: float = 1.0,
    ):
        """Initialize the one euro filter."""
        # Previous values.
        self.previous_value: float = initial_value
        self.previous_derivative: float = initial_derivative
        self.previous_time: float = initial_time
        # The parameters.
        self.min_cutoff: float = min_cutoff
        self.beta: float = beta
        self.derivative_cutoff: float = derivative_cutoff

    def smoothing_factor(
            self,
            time_elapsed: float,
            cutoff_frequency: float) -> float:
            r = 2 * math.pi * cutoff_frequency * time_elapsed
            return self.__no_nan(r) / (self.__no_nan(r) + 1)

    def exp_smoothing(
            self,
            alpha: float,
            current_value: float,
            previous_value: float
        ) -> float:
        return self.__no_nan(alpha) * self.__no_nan(current_value) + (1 - self.__no_nan(alpha)) * self.__no_nan(previous_value)
    
    def __no_nan(self, value):
        return value if not np.isnan(value) else 0.0

    def __call__(self, current_time: float, current_value: float) -> float:
        """Compute the filtered signal."""
        time_elapsed = current_time - self.previous_time
        LOGGER.debug(f"time elapsed: {time_elapsed}")

        # The filtered derivative of the signal.
        alpha_derivative = self.smoothing_factor(time_elapsed, self.derivative_cutoff)
        current_derivative = (current_value - self.previous_value) / time_elapsed
        filtered_derivative = self.exp_smoothing(alpha_derivative, current_derivative, self.previous_derivative)
        LOGGER.debug(f"alpha_derivative: {alpha_derivative}, current_derivative: {current_derivative}, filtered_derivative: {filtered_derivative}")

        # The filtered signal.
        adaptive_cutoff = self.min_cutoff + self.beta * abs(filtered_derivative)
        alpha = self.smoothing_factor(time_elapsed, adaptive_cutoff)
        filtered_value = self.exp_smoothing(alpha, current_value, self.previous_value)
        LOGGER.debug(f"alpha: {alpha}, value: {current_value}, previous_value: {self.previous_value}, filtered: {filtered_value}")

        # Memorize the previous values.
        self.previous_value = filtered_value
        self.previous_derivative = filtered_derivative
        self.previous_time = current_time

        return filtered_value
    
def get_euler_form(point, reference=None):
    "If reference provided, point of origin changes to reference. Provide the parameters in cartesian form"
    point = complex(point[0], point[1])
    if reference is not None:
        point = point - complex(reference[0] - reference[1])
    p = math.atan2(point.imag, point.real)
    m = math.sqrt(point.real**2 + point.imag**2)
    return m,p

def get_cartesian(euler, reference=None):
    x = euler[0] * math.cos(euler[1])
    y = euler[0] * math.sin(euler[1])
    if reference is not None:
        ref_x, ref_y = get_cartesian(reference)
        x = x + ref_x
        y = y + ref_y
    return x, y

def phase_to_degree(phase):
    if phase < 0:
        phase += 2 * math.pi
    return phase * 180 / math.pi

def degree_to_phase(degree):
    if degree < 0:
        degree += 360
    return degree * math.pi / 180

def get_actual_from_relative(relative, screen_width, screen_height, dtype='int'):
    pixel_x = relative[0]*screen_width
    pixel_y = relative[1]*screen_height
    return (pixel_x, pixel_y) if dtype == 'float' else (int(pixel_x), int(pixel_y))

def get_relative_from_actual(actual, screen_width, screen_height):
    pixel_x = actual[0]/screen_width
    pixel_y = actual[1]/screen_height
    return (pixel_x, pixel_y)

def get_distance(point1, point2):
    return math.sqrt((point1[0]-point2[0])**2 + (point1[1]-point2[1])**2)

def get_file_names(prefix, directory=None):
    '''
    if directory is None, check from the default path specified in EyeTrackerAnalyzer.__dirpath__
    '''
    if directory is None:
        directory = __datapath__
    directory = os.path.abspath(directory)
    if os.path.exists(directory):
        return glob.glob(os.path.join(directory, f'{prefix}*'))
    return []



import PyQt6.QtWidgets as qtw
import PyQt6.QtCore as qtc
import PyQt6.QtGui as qtg
import sys
import random
import datetime
import json
import threading
import os
import click
from typing import Union, Optional
from pyETA import __datapath__, LOGGER
import pyETA.components.utils as eta_utils
from pyETA.components.reader import TrackerThread

class ValidationWindow(qtw.QMainWindow):
    def __init__(self):
        super().__init__()
        self.total_grids = (3,3)
        self.circle_positions = [(row, col) for row in range(self.total_grids[0]) for col in range(self.total_grids[1])]
        self.current_position = None
        self.movement_duration = 1000
        self.stay_duration = 3000
        self.circle_size = 20
        self.collected_data = []
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Validation Window')
        #self.showFullScreen()
        self.screen_width, self.screen_height = self.size().width(), self.size().height()
        LOGGER.info(f"Screen width x height: {self.screen_width} x {self.screen_height}")
        self.gridWidget = qtw.QWidget(self)
        self.setCentralWidget(self.gridWidget)

        layout = qtw.QGridLayout(self.gridWidget)

        for row in range(self.total_grids[0]):
            for col in range(self.total_grids[1]):
                label = qtw.QLabel('+', self)
                label.setAlignment(qtc.Qt.AlignmentFlag.AlignCenter)
                layout.addWidget(label, row, col)

        self.circle = qtw.QLabel(self)
        self.circle.setStyleSheet("background-color: blue; border-radius: 10px;")
        self.circle.setFixedSize(self.circle_size, self.circle_size)
        self.circle.hide()

        self.show()
        self.animation = qtc.QPropertyAnimation(self.circle, b"pos")
        self.animation.finished.connect(self.on_animation_finished)
        qtc.QTimer.singleShot(self.stay_duration*3, self.start_sequence)

    def start_sequence(self):
        if self.circle_positions:
            self.move_to_next_position()
        else:
            self.circle.hide()
            self.process_data()
            LOGGER.info("Sequence completed!")
            qtc.QTimer.singleShot(self.stay_duration*3, self.close)

    def move_to_next_position(self):
        if not self.circle_positions:
            return

        next_position = random.choice(self.circle_positions)
        self.current_position = next_position
        target_widget = self.gridWidget.layout().itemAtPosition(*next_position).widget()
        target_pos = target_widget.mapTo(self, qtc.QPoint(target_widget.width() // 2 - self.circle_size // 2,
                                                          target_widget.height() // 2 - self.circle_size // 2))

        self.current_target_pos = target_pos
        self.animation.setStartValue(self.circle.pos())
        self.animation.setEndValue(target_pos)
        self.animation.setDuration(self.movement_duration)
        
        self.circle.show()
        self.animation.start()

        self.circle_positions.remove(self.current_position)

    def on_animation_finished(self):
        self.collect_data()
        qtc.QTimer.singleShot(self.stay_duration, self.start_sequence)

    def keyPressEvent(self, event: qtg.QKeyEvent):
        if event.key() == qtc.Qt.Key.Key_F11 or event.key() == qtc.Qt.Key.Key_F:
            self.showNormal() if self.isFullScreen() else self.showFullScreen()
            self.screen_width, self.screen_height = self.size().width(), self.size().height()
            LOGGER.info(f"Screen width x height: {self.screen_width} x {self.screen_height}")
        elif event.key() == qtc.Qt.Key.Key_Escape or event.key() == qtc.Qt.Key.Key_Q:
            LOGGER.info("Validation Window closed manually!")
            self.close()
        else:
            super().keyPressEvent(event)
    
    def collect_data(self):
        circle_center = qtc.QPoint(self.circle_size // 2, self.circle_size // 2)

        window_pos = self.circle.mapTo(self, circle_center)
        relative_pos = eta_utils.get_relative_from_actual((window_pos.x(), window_pos.y()), self.width(), self.height())
        recalibrated_pos = eta_utils.get_actual_from_relative(relative_pos, self.screen_width, self.screen_height)
        data_point = {
            "timestamp": eta_utils.get_timestamp(),
            "grid_position": self.current_position,
            "screen_position": recalibrated_pos
        }
        LOGGER.debug(f"Grid: {data_point.get('grid_position')}, Target: {self.current_target_pos}, Screen   : {data_point.get('screen_position')}")
        self.collected_data.append(data_point)

    def process_data(self):
        if not os.path.exists(__datapath__):
            os.makedirs(__datapath__)
        file = os.path.join(__datapath__, f"system_{eta_utils.get_system_info()}.json")
        with open(file, "w") as f:
            json.dump(
                {
                    "screen_size": (self.screen_width, self.screen_height),
                    "stay_duration": self.stay_duration,
                    "data": self.collected_data
                }, f, indent=4)
            LOGGER.info(f"Validation Data saved: {file}!")

def run_validation_window(screen_index: Optional[int] = 0):
    validation_window = ValidationWindow()
    screens = qtw.QApplication.screens()

    if screen_index < len(screens):
        screen = screens[screen_index]
        geometry = screen.availableGeometry()
        validation_window.setGeometry(geometry)
        validation_window.move(geometry.topLeft())
        LOGGER.info(f"Validation Window created on screen {screen_index + 1} with resolution: {geometry.width()}x{geometry.height()}")
    else:
        raise ValueError(f"Invalid screen index: {screen_index}")
    return validation_window

@click.command(name="window")
@click.option("--use_mock", is_flag=True, help="Use mockup tracker")
@click.option("--screen_index", default=0, help="Screen index to display the validation window")
@click.option("--verbose", is_flag=True, help="Enable verbose logging")
def main(use_mock, screen_index, verbose):
    app = qtw.QApplication(sys.argv)
    validation_window = run_validation_window(screen_index=screen_index)
    tracker_params = {
        'use_mock': use_mock,
        'fixation': False,
        'verbose': verbose,
        'push_stream': False,
        'save_data': True,
        'screen_index': screen_index,
        'duration': (9*(3000+1000))/1000 + (2000*3)/1000 + 2000/1000
    }


    tracker_thread = TrackerThread()
    tracker_thread.set_variables(tracker_params)
    tracker_thread.finished_signal.connect(lambda msg: LOGGER.info(msg))
    tracker_thread.error_signal.connect(lambda msg: LOGGER.error(msg))
    tracker_thread.start()
    validation_window.show()
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


import sys
import click
import psutil
import os
import datetime
import threading
import numpy as np
import pandas as pd
import pyqtgraph as pg

import PyQt6.QtWidgets as qtw
import PyQt6.QtCore as qtc
import PyQt6.QtGui as qtg
from typing import Optional

from pyETA import __version__, LOGGER, __datapath__

from pyETA.components.reader import StreamThread, TrackerThread
import pyETA.components.utils as eta_utils
import pyETA.components.validate as eta_validate


class EyeTrackerAnalyzer(qtw.QMainWindow):
    """Main application window for pyETA"""
    def __init__(self):
        super().__init__()
        self.setWindowTitle(f"pyETA-{__version__}")
        self.resize(1200, 800)
        self.stream_thread = None
        self.validate_thread = None
        self.is_gaze_playing = False
        self.is_fixation_playing = False
        self.start_time = None

        self.plot_timer = qtc.QTimer()
        self.plot_timer.timeout.connect(self.update_plots_from_stream)
        self.plot_timer.start(500)

        # Central widget and main layout
        central_widget = qtw.QWidget()
        self.setCentralWidget(central_widget)

        # Splitter for collapsible sidebar
        splitter = qtw.QSplitter(qtc.Qt.Orientation.Horizontal, central_widget)

        # Sidebar
        self.sidebar = self.create_sidebar()
        splitter.addWidget(self.sidebar)
        self.system_info_timer = qtc.QTimer()
        self.system_info_timer.timeout.connect(self.update_system_info)
        self.system_info_timer.start(1000)

        # Main content area
        main_content_widget = qtw.QWidget()
        main_layout = qtw.QVBoxLayout(main_content_widget)

        # Stream Configuration and System Info
        config_info_layout = qtw.QVBoxLayout()
        stream_config_layout = self.create_stream_configuration()
        config_info_layout.addLayout(stream_config_layout)
        line = qtw.QFrame()
        line.setFrameShape(qtw.QFrame.Shape.HLine)
        config_info_layout.addWidget(line)
        main_layout.addLayout(config_info_layout)

        # Tabs
        self.tab_widget = qtw.QTabWidget()
        main_layout.addWidget(self.tab_widget)
        self.setup_tabs()

        splitter.addWidget(main_content_widget)
        splitter.setStretchFactor(0, 1)
        splitter.setStretchFactor(1, 4)

        # Finalize layout
        layout = qtw.QVBoxLayout(central_widget)
        layout.addWidget(splitter)
        self.update_status_bar("pyETA status OK", 1, 5000)

        self.setStyleSheet("""
            QPushButton:hover {
                border: 1px solid black;
                border-radius: 5px;
                background-color: #2ECC71; 
            }
        """)

    def update_status_bar(self, message, state=3, timeout=5000):
        """
        Updates the status bar with the given message and state.

        0: Error (Red)
        1: Success (Green)
        2: Processing (Yellow)
        3: Default (None)
        """
        if state == 2:
            self.statusBar().setStyleSheet("background-color: #FFFF00; color: black;")
        elif state == 0:
            self.statusBar().setStyleSheet("background-color: #FF0000; color: white;")
        elif state == 1:
            self.statusBar().setStyleSheet("background-color: #2ECC71; color: black;")
        else:
            self.statusBar().setStyleSheet("background-color: none;")
        self.statusBar().showMessage(message, timeout)
        qtc.QTimer.singleShot(timeout, lambda: self.statusBar().setStyleSheet("background-color: none;"))

    def create_sidebar(self):
        frame = qtw.QFrame()
        layout = qtw.QVBoxLayout(frame)

        title = qtw.QLabel("<h1>Toolbox - Eye Tracker Analyzer</h1>")
        title.setStyleSheet("color: gray; margin-bottom: 10px;")
        layout.addWidget(title)

        self.source_code_link = qtw.QTextBrowser()
        self.source_code_link.setOpenExternalLinks(True)
        self.source_code_link.setReadOnly(True)
        self.source_code_link.setHtml(
            """
            <h3 style='color: #555;'>Faculty 1</h3>
            <a href='https://www.b-tu.de/en/fg-neuroadaptive-hci/' style='text-decoration: none;'>
                <strong>Neuroadaptive Human-Computer Interaction</strong><br>
                Brandenburg University of Technology (Cottbus-Senftenberg)
            </a>
            <h3 style='color: #555;'>Source code</h3>
            <a href='https://github.com/VinayIN/EyeTrackerAnalyzer.git' style='text-decoration: none;' target='_blank'>
                https://github.com/VinayIN/EyeTrackerAnalyzer.git
            </a>
            <h3 style='color: #555;'>Documentation</h3>
            <a href='https://vinayin.gitbook.io/pyeta/' style='text-decoration: none;' target='_blank'>
                https://vinayin.gitbook.io/pyeta
            """
        )
        self.source_code_link.setStyleSheet("margin-bottom: 20px; background: transparent;")
        self.source_code_link.anchorClicked.connect(lambda url: qtg.QDesktopServices.openUrl(url))

        markdown_text = qtw.QLabel(
            f"""<p>pyETA, Version: <code>{__version__}</code></p>
            <p>This interface allows you to validate the eye tracker accuracy along with the following:</p>
            <ul>
                <li>View gaze points</li>
                <li>View fixation points</li>
                <li>View eye tracker accuracy</li>
                <ul><li>Comparing the gaze data with validation grid locations.</li></ul>
            </ul>"""
        )
        layout.addWidget(markdown_text)
        layout.addWidget(self.source_code_link)

        self.system_info_card = self.create_system_info_card()
        layout.addWidget(self.system_info_card)
        return frame

    def create_system_info_card(self):
        card = qtw.QFrame()
        card.setFrameShape(qtw.QFrame.Shape.Box)
        layout = qtw.QVBoxLayout(card)
        system_buttons = qtw.QHBoxLayout()
        refresh_button = qtw.QPushButton("Refresh application")
        exit_button = qtw.QPushButton("Exit")
        exit_button.clicked.connect(self.close)
        refresh_button.clicked.connect(self.refresh_application)
        system_buttons.addWidget(exit_button)
        system_buttons.addWidget(refresh_button)
        layout.addLayout(system_buttons)

        # Add refresh slider for plot updates
        refresh_rate_layout = qtw.QHBoxLayout()
        refresh_rate_label = qtw.QLabel("Refresh Rate (ms):")
        self.refresh_slider = qtw.QSlider(qtc.Qt.Orientation.Horizontal)
        self.refresh_slider.setMinimum(10)
        self.refresh_slider.setMaximum(1000)
        self.refresh_slider.setValue(200)
        self.refresh_label = qtw.QLabel("200 ms")
        self.refresh_slider.valueChanged.connect(self.update_plot_refresh_rate)
        refresh_rate_layout.addWidget(refresh_rate_label)
        refresh_rate_layout.addWidget(self.refresh_slider)
        refresh_rate_layout.addWidget(self.refresh_label)
        layout.addLayout(refresh_rate_layout)

        self.system_info_labels = {
            "status": qtw.QLabel(),
            "pid": qtw.QLabel(),
            "stream id": qtw.QLabel(),
            "validate id": qtw.QLabel(),
            "total threads": qtw.QLabel(),
            "runtime": qtw.QLabel(),
            "memory": qtw.QLabel(),
            "storage": qtw.QLabel(),
            "cpu": qtw.QLabel(),
        }

        for label_name, label in self.system_info_labels.items():
            layout.addWidget(label)

        self.update_system_info()
        return card

    def update_plot_refresh_rate(self, value):
        self.plot_timer.setInterval(value)
        self.refresh_label.setText(f"{value} ms")
        LOGGER.info(f"Plot refresh rate set to {value} ms")

    def update_system_info(self):
        process = psutil.Process(os.getpid())
        memory_info = process.memory_info()
        cpu_percent = process.cpu_percent(interval=0.1)
        storage_free = psutil.disk_usage(os.getcwd()).free / 1024**3  # Free storage in GB
        runtime = datetime.datetime.now() - datetime.datetime.fromtimestamp(process.create_time())

        self.system_info_labels["pid"].setText(f"<strong>Application PID:</strong> {process.pid}")
        self.system_info_labels["stream id"].setText(
            f"<strong>Stream Thread ID:</strong> {self.stream_thread.id if self.stream_thread else 'Not Running'}"
        )
        self.system_info_labels["validate id"].setText(
            f"<strong>Validate Thread ID:</strong> {self.validate_thread.id if self.validate_thread else 'Not Running'}"
        )
        self.system_info_labels["total threads"].setText(f"<strong>Total Threads:</strong> {threading.active_count()}")
        self.system_info_labels["runtime"].setText(f"<strong>Runtime:</strong> {runtime}")
        self.system_info_labels["memory"].setText(f"<strong>Memory:</strong> {memory_info.rss / 1024**2:.1f} MB")
        self.system_info_labels["storage"].setText(f"<strong>Storage available:</strong> {storage_free:.1f} GB")
        self.system_info_labels["cpu"].setText(f"<strong>CPU Usage:</strong> {cpu_percent:.1f}%")

    def refresh_application(self):
        self.source_code_link.clearFocus()
        self.gaze_plot_x_curve.setData([], [])
        self.gaze_plot_y_curve.setData([], [])
        self.fixation_scatter.setData([], [])
        self.update_metric_tab()
        self.metrics_table.clear()
        self.update_status_bar("Application refreshed successfully", 1, 5000)
    
    def create_stream_configuration(self):
        main_layout = qtw.QHBoxLayout()
        layout_first = qtw.QVBoxLayout()

        stream_type_layout = qtw.QHBoxLayout()
        stream_type_label = qtw.QLabel("Stream Type:")
        self.stream_type_combo = qtw.QComboBox()
        self.stream_type_combo.addItems(["Eye-Tracker", "Mock"])
        stream_type_layout.addWidget(stream_type_label)
        stream_type_layout.addWidget(self.stream_type_combo)
        layout_first.addLayout(stream_type_layout)

        data_rate_layout = qtw.QHBoxLayout()
        data_rate_label = qtw.QLabel("Data Rate (Hz):")
        self.data_rate_slider = qtw.QSlider(qtc.Qt.Orientation.Horizontal)
        self.data_rate_slider.setMinimum(0)
        self.data_rate_slider.setMaximum(800)
        self.data_rate_slider.setValue(600)
        self.data_rate_label = qtw.QLabel("600 Hz")
        self.data_rate_slider.valueChanged.connect(lambda value: self.data_rate_label.setText(f"{value} Hz"))
        data_rate_layout.addWidget(data_rate_label)
        data_rate_layout.addWidget(self.data_rate_slider)
        data_rate_layout.addWidget(self.data_rate_label)
        layout_first.addLayout(data_rate_layout)

        velocity_threshold_layout = qtw.QHBoxLayout()
        velocity_threshold_label = qtw.QLabel("Velocity Threshold:")
        self.velocity_threshold_spinbox = qtw.QDoubleSpinBox()
        self.velocity_threshold_spinbox.setRange(0.0, 50.0)
        self.velocity_threshold_spinbox.setValue(5.0)
        self.velocity_threshold_spinbox.setSingleStep(0.1)
        self.velocity_threshold_spinbox.valueChanged.connect(
            lambda value: self.velocity_threshold_label.setText(f"{value:.1f}")
        )
        self.velocity_threshold_label = qtw.QLabel("50.0")
        velocity_threshold_layout.addWidget(velocity_threshold_label)
        velocity_threshold_layout.addWidget(self.velocity_threshold_spinbox)
        velocity_threshold_layout.addWidget(self.velocity_threshold_label)
        layout_first.addLayout(velocity_threshold_layout)

        main_layout.addLayout(layout_first)

        layout_second = qtw.QVBoxLayout()
        self.fixation_check = qtw.QCheckBox("Enable Fixation")
        self.fixation_check.setChecked(True)
        layout_second.addWidget(self.fixation_check)

        self.push_stream_check = qtw.QCheckBox("Push to Stream")
        self.push_stream_check.setChecked(True)
        layout_second.addWidget(self.push_stream_check)

        self.verbose_check = qtw.QCheckBox("Verbose Mode")
        layout_second.addWidget(self.verbose_check)

        self.dont_screen_nans_check = qtw.QCheckBox("Accept Screen NaNs (Default: 0)")
        layout_second.addWidget(self.dont_screen_nans_check)

        main_layout.addLayout(layout_second)

        control_layout = qtw.QVBoxLayout()
        start_stop_layout = qtw.QHBoxLayout()
        self.start_stream_btn = qtw.QPushButton("Start Stream")
        self.stop_stream_btn = qtw.QPushButton("Stop Stream")
        self.start_stream_btn.clicked.connect(self.start_stream)
        self.stop_stream_btn.clicked.connect(self.stop_stream)
        start_stop_layout.addWidget(self.start_stream_btn)
        start_stop_layout.addWidget(self.stop_stream_btn)
        control_layout.addLayout(start_stop_layout)

        self.validate_btn = qtw.QPushButton("Validate Eye Tracker")
        self.validate_btn.clicked.connect(self.validate_eye_tracker)
        control_layout.addWidget(self.validate_btn)

        main_layout.addLayout(control_layout)
        return main_layout

    def setup_tabs(self):
        self.gaze_tab = self.create_gaze_data_tab()
        self.fixation_tab = self.create_fixation_tab()
        self.metrics_tab = self.create_metrics_tab()

        self.tab_widget.addTab(self.gaze_tab, "Gaze Data")
        self.tab_widget.addTab(self.fixation_tab, "Fixation")
        self.tab_widget.addTab(self.metrics_tab, "Metrics")

    def validate_eye_tracker(self):
        screen_dialog = qtw.QDialog()
        screen_dialog.setWindowTitle("Select Validation Screen")
        layout = qtw.QVBoxLayout(screen_dialog)

        screens = qtw.QApplication.screens()
        screen_combo = qtw.QComboBox()
        for i, screen in enumerate(screens):
            screen_combo.addItem(f"Screen {i+1}: {screen.geometry().width()}x{screen.geometry().height()}")

        layout.addWidget(qtw.QLabel("Choose Validation Screen:"))
        layout.addWidget(screen_combo)

        if self.validate_thread and self.validate_thread.isRunning():
            qtw.QMessageBox.warning(self, "Warning", "Validation Tracker is already running. Please stop the stream")
            return

        def start_validation():
            selected_screen_index = screen_combo.currentIndex()
            tracker_params = {
                'use_mock': self.stream_type_combo.currentText() == "Mock",
                'fixation': False,
                'verbose': self.verbose_check.isChecked(),
                'push_stream': False,
                'save_data': True,
                'screen_index': selected_screen_index,
                'duration': (9*(3000+1000))/1000 + (2000*3)/1000 + 2000/1000
            }

            try:
                from pyETA.components.window import run_validation_window

                self.validation_window = run_validation_window(screen_index=selected_screen_index)
                self.validate_thread = TrackerThread()
                self.validate_thread.set_variables(tracker_params)
                self.validate_thread.finished_signal.connect(
                    lambda msg: qtw.QMessageBox.information(self, "Validation Thread", msg)
                )
                self.validate_thread.error_signal.connect(
                    lambda msg: qtw.QMessageBox.critical(self, "Validation Thread", msg)
                )
                self.validate_thread.start()
                self.validation_window.show()

                self.update_status_bar("Validation started", 2, 10000)
                screen_dialog.close()

            except Exception as e:
                qtw.QMessageBox.critical(self, "Validation Error", str(e))
                LOGGER.error(f"Validation error: {str(e)}")

        validate_btn = qtw.QPushButton("Start Validation")
        validate_btn.clicked.connect(start_validation)
        layout.addWidget(validate_btn)
        screen_dialog.exec()
    
    def create_gaze_data_tab(self):
        tab = qtw.QWidget()
        layout = qtw.QVBoxLayout(tab)

        control_panel = qtw.QHBoxLayout()
        self.gaze_play_btn = qtw.QPushButton("Play")
        self.gaze_play_btn.setFixedSize(60, 30)
        self.gaze_play_btn.clicked.connect(self.toggle_gaze_play)
        self.gaze_stream_label = qtw.QLabel("Stream: Not Connected")
        control_panel.addWidget(self.gaze_play_btn)
        control_panel.addWidget(self.gaze_stream_label)
        layout.addLayout(control_panel)

        self.gaze_plot_x = pg.PlotWidget(title="Gaze X Position")
        self.gaze_plot_x.showGrid(x=True, y=True)
        self.gaze_plot_x.setYRange(0, self.screen().size().width())
        self.gaze_plot_x.setLabel('bottom', 'Time (s)')
        self.gaze_plot_x.setLabel('left', 'Pixel Position - Width')
        self.gaze_plot_x_curve = self.gaze_plot_x.plot(pen='y')
        layout.addWidget(self.gaze_plot_x)

        self.gaze_plot_y = pg.PlotWidget(title="Gaze Y Position")
        self.gaze_plot_y.showGrid(x=True, y=True)
        self.gaze_plot_y.setYRange(0, self.screen().size().height())
        self.gaze_plot_y.setLabel('bottom', 'Time (s)')
        self.gaze_plot_y.setLabel('left', 'Pixel Position - Height')
        self.gaze_plot_y_curve = self.gaze_plot_y.plot(pen='r')
        layout.addWidget(self.gaze_plot_y)
        
        return tab

    def create_fixation_tab(self):
        tab = qtw.QWidget()
        layout = qtw.QVBoxLayout(tab)

        control_panel = qtw.QHBoxLayout()
        self.fixation_play_btn = qtw.QPushButton("Play")
        self.fixation_play_btn.setFixedSize(60, 30)
        self.fixation_play_btn.clicked.connect(self.toggle_fixation_play)
        self.fixation_stream_label = qtw.QLabel("Stream: Not Connected")
        control_panel.addWidget(self.fixation_play_btn)
        control_panel.addWidget(self.fixation_stream_label)
        layout.addLayout(control_panel)

        self.fixation_plot = pg.PlotWidget(title="Fixation Points")
        self.fixation_plot.setXRange(0, self.screen().size().width())
        self.fixation_plot.setYRange(0, self.screen().size().height())
        self.fixation_plot.getAxis('left').setLabel('Pixel Position - Height')
        self.fixation_plot.getAxis('bottom').setLabel('Pixel Position - Width')
        self.fixation_plot.invertY(True)
        self.fixation_scatter = pg.ScatterPlotItem()
        self.fixation_plot.addItem(self.fixation_scatter)
        layout.addWidget(self.fixation_plot)
        return tab
    
    def toggle_gaze_play(self):
        self.is_gaze_playing = not self.is_gaze_playing
        self.gaze_play_btn.setText("Pause" if self.is_gaze_playing else "Play")
    
    def toggle_fixation_play(self):
        self.is_fixation_playing = not self.is_fixation_playing
        self.fixation_play_btn.setText("Pause" if self.is_fixation_playing else "Play")
    
    def get_gaze_and_validate_data(self):
        gaze = sorted(eta_utils.get_file_names(prefix="gaze_data_"))
        validate = sorted(eta_utils.get_file_names(prefix="system_"))
        return gaze, validate

    def create_metrics_tab(self):
        tab = qtw.QWidget()
        layout = qtw.QVBoxLayout(tab)
        metrics_title = qtw.QLabel("<h2>Statistics: Eye Tracker Validation</h2>")
        metrics_datapath = qtw.QLabel(f"Searching data files at path: {__datapath__}")
        file_selector = qtw.QHBoxLayout()
        
        self.gaze_data = qtw.QComboBox()
        self.validate_data = qtw.QComboBox()

        file_selector.addWidget(self.gaze_data)
        file_selector.addWidget(self.validate_data)
        self.update_metric_tab()

        validate_btn = qtw.QPushButton("Validate")
        self.df = pd.DataFrame()
        validate_btn.clicked.connect(self.update_metrics_table)

        layout.addWidget(metrics_title)
        layout.addWidget(metrics_datapath)
        layout.addLayout(file_selector)
        layout.addWidget(validate_btn)

        self.metrics_table = qtw.QTableWidget()
        layout.addWidget(self.metrics_table)
        download_btn = qtw.QPushButton("Download CSV")
        download_btn.clicked.connect(self.download_csv)
        layout.addWidget(download_btn)

        return tab

    def update_plots_from_stream(self):
        if not self.stream_thread or not self.stream_thread.isRunning() or not self.start_time:
            return

        if self.is_gaze_playing:
            gaze_data = self.stream_thread.get_data(fixation=False)
            if len(gaze_data) > 0:
                timestamps = gaze_data['timestamp']
                x_coord = gaze_data['x']
                y_coord = gaze_data['y']
                self.update_gaze_plot(timestamps, x_coord, y_coord)
    
        if self.is_fixation_playing:
            fixation_data = self.stream_thread.get_data(fixation=True)
            if len(fixation_data) > 0:
                x_coord = fixation_data['x']
                y_coord = fixation_data['y']
                count = fixation_data['count']
                timestamp = fixation_data['timestamp']
                self.update_fixation_plot(x_coord, y_coord, count, timestamp)

    def update_gaze_plot(self, timestamp, x_coord, y_coord):
        current_time = timestamp[-1] - self.start_time
        window_size = 10

        relative_times = timestamp - self.start_time
        mask = relative_times >= (current_time - window_size)
        filtered_times = relative_times[mask]
        filtered_x = x_coord[mask]
        filtered_y = y_coord[mask]

        self.gaze_plot_x_curve.setData(filtered_times, filtered_x)
        self.gaze_plot_y_curve.setData(filtered_times, filtered_y)
        self.gaze_plot_x.setXRange(max(0, current_time - window_size), current_time)
        self.gaze_plot_y.setXRange(max(0, current_time - window_size), current_time)

    def update_fixation_plot(self, x_coord, y_coord, counts, timestamp):
        self.fixation_scatter.setData(
            x=x_coord,
            y=y_coord,
            size = np.minimum(counts/10, 50),
            symbol='o'
        )

    def update_metric_tab(self):
        self.gaze_data_items, self.validate_data_items = self.get_gaze_and_validate_data()
        self.gaze_data.clear()
        self.gaze_data.addItems(['select gaze data'] + [
            f"File {idx+1}: {eta_validate.get_gaze_data_timestamp(file)}" for idx, file in enumerate(self.gaze_data_items)
        ])
        self.validate_data.clear()
        self.validate_data.addItems(['select validation data'] + [
            f"File {idx+1} {eta_validate.get_validate_data_timestamp(file)}" for idx, file in enumerate(self.validate_data_items)
        ])

    def start_stream(self):
        if self.stream_thread and self.stream_thread.isRunning():
            qtw.QMessageBox.warning(self, "Warning", "Stream is already running.")
            return
        
        tracker_params = {
            'data_rate': self.data_rate_slider.value(),
            'use_mock': self.stream_type_combo.currentText() == "Mock",
            'fixation': self.fixation_check.isChecked(),
            'velocity_threshold': self.velocity_threshold_spinbox.value(),
            'dont_screen_nans': self.dont_screen_nans_check.isChecked(),
            'verbose': self.verbose_check.isChecked(),
            'push_stream': self.push_stream_check.isChecked(),
            'save_data': False,
        }
        
        try:
            self.update_status_bar("Stream: Connecting...", 2, 2000)
            self.start_time = datetime.datetime.now().timestamp()
            self.stream_thread = StreamThread()
            self.stream_thread.set_variables(tracker_params=tracker_params)
            self.stream_thread.found_signal.connect(lambda msg: self.update_plot_label(msg))
            self.stream_thread.error_signal.connect(lambda msg: qtw.QMessageBox.critical(self, "Error", msg))
            self.stream_thread.start()
            self.plot_timer.start(self.refresh_slider.value())  # Start timer with current slider value

        except Exception as e:
            error_msg = f"Failed to start stream: {str(e)}"
            LOGGER.error(error_msg)
            self.update_status_bar(error_msg, 0, 5000)

    def stop_stream(self):
        if not self.stream_thread or not self.stream_thread.isRunning():
            qtw.QMessageBox.warning(self, "Warning", "No active stream to stop.")
            return
        
        try:
            self.stream_thread.stop()
            self.stream_thread = None
            self.start_time = None
            self.plot_timer.stop()  # Stop the plot update timer
            self.update_status_bar("Stream stopped successfully", 1, 3000)
            self.is_gaze_playing = False
            self.is_fixation_playing = False
            self.gaze_play_btn.setText("Play")
            self.fixation_play_btn.setText("Play")
            self.update_plot_label()
            LOGGER.warning(f"Thread count after stop stream: {threading.active_count()}")
        except Exception as e:
            LOGGER.error(f"Error stopping stream: {str(e)}")
            self.update_status_bar(f"Error stopping stream: {str(e)}", 0, 5000)

    def update_plot_label(self, msg="Stream: Not Connected"):
        self.gaze_stream_label.setText(msg)
        self.fixation_stream_label.setText(msg)
        if "✔" in msg:
            self.update_status_bar("Stream connected successfully", 1, 5000)
        if "✘" in msg:
            self.update_status_bar("Stream connection failed", 0, 5000)

    def update_metrics_table(self):
        self.update_status_bar("Calculating", 2, 5000)
        self.combined_df, df, described_df = eta_validate.get_statistics(
            gaze_file=self.gaze_data_items[self.gaze_data.currentIndex() - 1],
            validate_file=self.validate_data_items[self.validate_data.currentIndex() - 1]
        )
        LOGGER.info(f"Combined DataFrame: {self.combined_df.shape}, Grid DataFrame: {df.shape}, Described DataFrame: {described_df.shape}")

        self.metrics_table.setRowCount(self.combined_df.shape[0])
        self.metrics_table.setColumnCount(self.combined_df.shape[1])
        self.metrics_table.setHorizontalHeaderLabels(self.combined_df.columns)

        for row in range(self.combined_df.shape[0]):
            for col in range(self.combined_df.shape[1]):
                item = qtw.QTableWidgetItem(str(self.combined_df.iloc[row, col]))
                item.setTextAlignment(qtc.Qt.AlignmentFlag.AlignCenter)
                self.metrics_table.setItem(row, col, item)
        
        self.metrics_table.setAlternatingRowColors(True)
        self.metrics_table.resizeColumnsToContents()
        self.update_status_bar("Metrics generated successfully", 1, 8000)

    def download_csv(self):
        if self.combined_df.empty:
            qtw.QMessageBox.critical(self, "Error", "No data to save as CSV")
            return
        
        filename, _ = qtw.QFileDialog.getSaveFileName(self, "Save CSV", "", "CSV Files (*.csv)")
        if filename:
            self.combined_df.to_csv(filename, index=False)
            self.update_status_bar(f"csv saved at: {os.path.abspath(filename)}", 1, 5000)

    def closeEvent(self, event):
        LOGGER.info("close event invoked.")
        self.system_info_timer.stop()
        if self.stream_thread and self.stream_thread.isRunning():
            self.stream_thread.stop()
            LOGGER.info("Stopping stream thread during closeEvent")
        if self.validate_thread and self.validate_thread.isRunning():
            self.validate_thread.stop()
            LOGGER.info("Stopping validate thread during closeEvent")
        LOGGER.warning(f"Threads alive: {[t.name for t in threading.enumerate()]}")
        event.accept()


@click.command(name="application")
def main():
    app = qtw.QApplication(sys.argv)
    window = EyeTrackerAnalyzer()
    window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()


import datetime
import random
import argparse
import time
from threading import Thread
from pynput import mouse
from pyETA.components import utils

EYETRACKER_GAZE_DATA = 'mock_gaze_data'
CAPABILITY_HAS_GAZE_DATA = 'mock_has_gaze_data'

KNOWN_DATA_IDS = [EYETRACKER_GAZE_DATA]
device_capabilities = [CAPABILITY_HAS_GAZE_DATA]

class MockEyeTracker(Thread):
    def __init__(self, data_rate=600, verbose=False):
        Thread.__init__(self, name="MockEyeTracker", daemon=True)
        self.verbose = verbose
        self.data_rate = data_rate
        self.screen_width, self.screen_height = utils.get_current_screen_size()
        self.address ="ZA03046BINAY2024"
        self.model = "ZA03046BINAY2024"
        self.device_name = "Mock Tracker"
        self.serial_number = "ZA03046BINAY2024"
        self.callbacks = dict()
        self.curr_x = 0.0
        self.curr_y = 0.0
        self.should_stop = False
        self.listener = None

    def subscribe_to(self, id, callback, as_dictionary=True):
        if id not in KNOWN_DATA_IDS:
            raise ValueError(f"Unknown data type: {id}")

        if callback is None:
            raise ValueError("Callback cannot be None")

        self.callbacks[id] = callback

    def unsubscribe_from(self, id, callback):
        if id not in KNOWN_DATA_IDS:
            raise ValueError(f"Unknown data type: {id}")

        if callback is None:
            raise ValueError("Callback cannot be None")
        
        if id in self.callbacks:
            del self.callbacks[id]

        if not self.callbacks:
            self.stop()

    def on_move(self, x, y):
        corr_x = max(0, min(x, self.screen_width - 1))
        corr_y = max(0, min(y, self.screen_height - 1))
        self.curr_x = corr_x / self.screen_width
        self.curr_y = corr_y / self.screen_height

    def run(self):
        self.listener = mouse.Listener(on_move=self.on_move, daemon=True)
        self.listener.start()
        while not self.should_stop:
            time.sleep(.99 / self.data_rate)
            x = self.curr_x
            y = self.curr_y

            if EYETRACKER_GAZE_DATA in self.callbacks:
                self.callbacks[EYETRACKER_GAZE_DATA](
                    {
                        "device_time_stamp": utils.get_timestamp(),
                        "system_time_stamp": utils.get_timestamp(),
                        'left_gaze_point_on_display_area': [x, y],
                        'left_gaze_point_validity': random.uniform(0, 1) > 0.5,
                        'right_gaze_point_on_display_area': [x, y],
                        'right_gaze_point_validity': random.uniform(0, 1) > 0.5,
                        'left_pupil_diameter': 8.0 + 4 * random.uniform(-1, 1),
                        'left_pupil_validity': random.uniform(0, 1) > 0.5,
                        'right_pupil_diameter': 8.0 + 4 * random.uniform(-1, 1),
                        'right_pupil_validity': random.uniform(0, 1) > 0.5
                    })
        self.listener.stop()
        self.listener.join()

    def stop(self):
        """Stop the mock eye tracker"""
        self.should_stop = True
        if self.listener and self.listener.is_alive():
            self.listener.stop()
            self.listener.join()
        self.callbacks.clear()
        self.join()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_rate", help="Data rate in Hz", type=int, default=60)
    parser.add_argument("--verbose", help="Print the gaze data", action="store_true")
    args = parser.parse_args()

    mock_eye_tracker = MockEyeTracker(data_rate=args.data_rate, verbose=args.verbose)
    mock_eye_tracker.start()
    while True:
        try:
            continue
        except KeyboardInterrupt:
            mock_eye_tracker.stop()


from importlib.metadata import version
import logging
import os

__version__ = version("pyETA-toolbox")
__datapath__ = os.path.join(os.getcwd(), 'eta_data')
CONSOLE_LOG_FORMAT = '%(asctime)s :: %(filename)s:%(lineno)d :: %(levelname)s :: %(message)s'
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
console_handler.setFormatter(logging.Formatter(CONSOLE_LOG_FORMAT))

LOGGER.addHandler(console_handler)

__all__ = [
    'LOGGER',
    '__version__',
    '__datapath__'
]
LOGGER.debug(f"pyETA version: {__version__}")
LOGGER.debug(f"Data path: {__datapath__}")


import pandas as pd
import numpy as np
from typing import Dict, List, Tuple, Optional
import json
import os
import sys
import click
import re
import datetime
import pyETA.components.utils as eta_utils
from pyETA import LOGGER, __datapath__
from dataclasses import dataclass

@dataclass
class ValidateData:
    screen: Tuple[int, int]
    window: Tuple[int, int]


def convert_window_to_screen_coordinates(point: Tuple[float, float], dtype="int") -> Tuple[float, float]:
    x = (point[0] / ValidateData.window[0]) * ValidateData.screen[0]
    y = (point[1] / ValidateData.window[1]) * ValidateData.screen[1]
    return (x, y) if dtype == "float" else (int(x), int(y))

def convert_screen_to_window_coordinates(point: Tuple[float, float], dtype="int") -> Tuple[float, float]:
    x = (point[0] / ValidateData.screen[0]) * ValidateData.window[0]
    y = (point[1] / ValidateData.screen[1]) * ValidateData.window[1]
    return (x, y) if dtype == "float" else (int(x), int(y))

def calculate_statistics(df: pd.DataFrame) -> pd.DataFrame:
    target = df.screen_position_recalibrated.iloc[0]
    mean_data_x= int((df["left_eye_x"].mean() + df["right_eye_x"].mean())/2)
    mean_data_y= int((df["left_eye_y"].mean() + df["right_eye_y"].mean())/2)
    result = {
        "group": df.group.iloc[0],
        "target position": target,
        "distance (target to gaze points)": (df["distance_left_from_target"].mean() + df["distance_right_from_target"].mean())/2,
        "spread (target to gaze points)": (df["distance_left_from_target"].std() + df["distance_right_from_target"].std())/2,
        "mean gaze point": (mean_data_x, mean_data_y),
        "distance (mean to target)": eta_utils.get_distance((mean_data_x, mean_data_y), target),
        "distance (mean to gaze points)": (df["distance_left_from_mean"].mean() + df["distance_right_from_mean"].mean())/2,
        "spread (mean to gaze points)": (df["distance_left_from_mean"].std() + df["distance_right_from_mean"].std())/2,
    }
    return pd.DataFrame(data = [result])

def load_data(validate_file: str, gaze_file: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    with open(validate_file, 'r') as f:
        validate_data = json.load(f)
    with open(gaze_file, 'r') as f:
        gaze_data = json.load(f)
    
    ValidateData.screen = gaze_data["screen_size"]
    ValidateData.window = validate_data["screen_size"]
    df_gaze_data = pd.DataFrame(data=gaze_data["data"]).fillna(0.0)
    
    df_validate_data = pd.DataFrame(data=validate_data["data"]).fillna(0.0)
    df_validate_data.index.name = 'group'
    df_validate_data = df_validate_data.reset_index()
    data = {"gaze": df_gaze_data, "validate": df_validate_data}
    LOGGER.info(f"Screen resolution: {ValidateData.screen}")
    LOGGER.info(f"Window resolution: {ValidateData.window}")
    return data

def preprocess_data(df_gaze_data: pd.DataFrame, df_validate_data: pd.DataFrame) -> pd.DataFrame:
    df_gaze_data['timestamp_0'] = pd.to_datetime(df_gaze_data['timestamp'], unit='s')
    
    df_validate_data['timestamp_0'] = pd.to_datetime(df_validate_data['timestamp'], unit='s')
    df_validate_data['timestamp_3'] = df_validate_data['timestamp_0'] + datetime.timedelta(seconds=3)

    df_validate_data["screen_position_recalibrated"] = df_validate_data.screen_position.apply(
        lambda x: convert_window_to_screen_coordinates(x))
    
    return df_gaze_data, df_validate_data

def filter_and_group_data(df_gaze_data: pd.DataFrame, df_validate_data: pd.DataFrame) -> pd.DataFrame:
    groups = []
    
    for _, row in df_validate_data.iterrows():
        filtered_gaze_data = df_gaze_data[
            (df_gaze_data['timestamp_0'] >= row['timestamp_0']) & 
            (df_gaze_data['timestamp_0'] <= row['timestamp_3'])
        ].copy()
        
        if not filtered_gaze_data.empty:
            filtered_gaze_data["group"] = row["group"]
            groups.append(filtered_gaze_data)
    
    return pd.concat(groups, ignore_index=True)

def calculate_data(df: pd.DataFrame) -> pd.DataFrame:
    target = df.screen_position_recalibrated.iloc[0]
    
    def extract_distance_target(eye_data):
        if np.isnan(eye_data["gaze_point"][0]):
            return pd.Series({'distance': np.nan, 'x': np.nan, 'y': np.nan})
        point = eta_utils.get_actual_from_relative(eye_data["gaze_point"], ValidateData.screen[0], ValidateData.screen[1])
        distance_pixel = eta_utils.get_distance(point, target)
        return pd.Series({'distance': distance_pixel, 'x': point[0], 'y': point[1]})
    
    left_eye_data = df.left_eye.apply(extract_distance_target)
    right_eye_data = df.right_eye.apply(extract_distance_target)

    df["left_eye_x"] = left_eye_data['x']
    df["left_eye_y"] = left_eye_data['y']
    df["right_eye_x"] = right_eye_data['x']
    df["right_eye_y"] = right_eye_data['y']
    mean_xy = (left_eye_data['x'].mean() + right_eye_data['x'].mean()) / 2, (left_eye_data['y'].mean() + right_eye_data['y'].mean()) / 2
    df["distance_left_from_target"] = left_eye_data['distance']
    df["distance_right_from_target"] = right_eye_data['distance']

    def extract_distance_mean(eye_data, mean_xy):
        if np.isnan(eye_data["gaze_point"][0]):
            return pd.Series({'distance': np.nan, 'x': np.nan, 'y': np.nan})
        point = eta_utils.get_actual_from_relative(eye_data["gaze_point"], ValidateData.screen[0], ValidateData.screen[1])
        distance = eta_utils.get_distance(point, mean_xy)
        return pd.Series({'distance': distance, 'x': point[0], 'y': point[1]})
    
    left_eye_data = df.left_eye.apply(lambda x: extract_distance_mean(x, mean_xy))
    right_eye_data = df.right_eye.apply(lambda x: extract_distance_mean(x, mean_xy))

    df["distance_left_from_mean"] = left_eye_data['distance']
    df["distance_right_from_mean"] = right_eye_data['distance']
    return df

def get_statistics(gaze_file: str, validate_file: str) -> pd.DataFrame:
    data = load_data(validate_file, gaze_file)
    df_gaze_data, df_validate_data = data.get("gaze"), data.get("validate")
    if df_gaze_data.empty or df_validate_data.empty:
        return pd.DataFrame()
    
    df_gaze_data, df_validate_data = preprocess_data(df_gaze_data, df_validate_data)
    df_groups = filter_and_group_data(df_gaze_data, df_validate_data)
    
    df_groups = df_groups.join(df_validate_data[["screen_position", "screen_position_recalibrated"]], on=["group"], how="left").dropna()
    
    df_calculated = df_groups.groupby("group").apply(calculate_data)
    
    statistics = df_calculated.reset_index(drop=True).groupby("group").apply(calculate_statistics).reset_index(drop=True)
    
    result = statistics.round(4)
    described = result.describe(percentiles=[0.25, 0.50, 0.75]).loc[['25%', '50%', '75%', 'mean']].drop(columns=['group'])
    described.index.name= 'group'
    described = described.reset_index().round(4)
    
    separator_row = pd.DataFrame([["---"] * result.shape[1]], columns=result.columns)
    combined = pd.concat([result, separator_row, described], ignore_index=True).fillna("")
    return combined, result, described

def get_gaze_data_timestamp(file: str) -> Optional[datetime.datetime]:
    try:
        timestamp_data = re.search(r"gaze_data_(.*).json", file).group(1)
        return datetime.datetime.strptime(timestamp_data, "%Y%m%d_%H%M%S")
    except AttributeError:
        LOGGER.error(f"Invalid file name: {file}")
        return None

def get_validate_data_timestamp(file: str) -> Optional[datetime.datetime]:
    try:
        info = re.search(r"system_(.*).json", file).group(1)
        info = info.split("_")
        return datetime.datetime.strptime("_".join(info[-2:]), "%Y%m%d_%H%M%S")
    except AttributeError:
        LOGGER.error(f"Invalid file name: {file}")
        return None

@click.command("validate")
@click.option('--csv', help="Name of the output CSV file.")
def main(csv: Optional[str]):
    # List files in __datapath__
    gaze_files = sorted(eta_utils.get_file_names(prefix="gaze_data_"))
    validate_files = sorted(eta_utils.get_file_names(prefix="system_"))

    # Display gaze files
    print("Available Gaze Data:")
    for i, file in enumerate(gaze_files):
        print(f'{i + 1}. {file} (Timestamp: {get_gaze_data_timestamp(file)})')
    try:
        gaze_choice = int(input("Enter a number: ")) - 1
        gaze_file = os.path.join(__datapath__, gaze_files[gaze_choice])
    except (ValueError, IndexError):
        LOGGER.error("Invalid input. Exiting...")
        sys.exit(1)

    # Display validate files
    print("\nAvailable Validate Data:")
    for i, file in enumerate(validate_files):
        print(f'{i + 1}. {file} (Timestamp: {get_validate_data_timestamp(file)})')
    try:
        validate_choice = int(input("Enter a number: ")) - 1
        validate_file = os.path.join(__datapath__, validate_files[validate_choice])
    except (ValueError, IndexError):
        LOGGER.error("Invalid input. Exiting...")
        sys.exit(1)

    # Process and save to CSV
    combined, result, described = get_statistics(gaze_file=gaze_file, validate_file=validate_file)
    if csv is None:
        print(described)
    combined.to_csv(csv, index=False)
    print(f"Results saved to {csv}")

if __name__ == "__main__":
    main()


import threading
from collections import deque
from mne_lsl import lsl

from pyETA.components.track import Tracker
from pyETA import LOGGER
import PyQt6.QtCore as qtc
import numpy as np


class TrackerThread(qtc.QThread):
    finished_signal = qtc.pyqtSignal(str)
    error_signal = qtc.pyqtSignal(str)

    def __init__(self):
        super().__init__()
        self.tracker = None
        self.running = False
        self.id = None
    
    def set_variables(self, tracker_params):
        self.tracker_params = tracker_params

    def run(self):
        try:
            self.running = True
            self.id = threading.get_native_id()
            LOGGER.info("Starting tracker thread...")
            self.tracker = Tracker(**self.tracker_params)
            self.tracker.start_tracking(duration=self.tracker_params.get('duration', None))
            self.finished_signal.emit("Tracking completed successfully")
        except Exception as e:
            error_msg = f"Tracker error: {str(e)}"
            LOGGER.error(error_msg)
            self.error_signal.emit(error_msg)

    def stop(self):
        """Stop tracker thread."""
        self.running = False
        self.id = None
        if self.tracker and self.tracker.id:
            self.tracker.signal_break()
            self.tracker = None
        self.wait()
        self.quit()
        LOGGER.info("Tracker thread stopped!")

class StreamThread(qtc.QThread):
    found_signal = qtc.pyqtSignal(str)
    error_signal = qtc.pyqtSignal(str)

    def __init__(self):
        super().__init__()
        self.running = False
        self.id = None
        self.buffer = deque(maxlen=10000)
        self.fixation_buffer = deque(maxlen=10)
        self.current_fixation = None
    
    def set_variables(self, tracker_params):
        self.tracker_thread = TrackerThread()
        self.tracker_thread.set_variables(tracker_params)
        self.tracker_thread.finished_signal.connect(lambda msg: LOGGER.info(msg))
        self.tracker_thread.error_signal.connect(self.error_signal.emit)

    def run(self):
        try:
            self.tracker_thread.start()
            streams = lsl.resolve_streams(timeout=1, name='tobii_gaze_fixation')
            if not streams:
                error_msg = "'tobii_gaze_fixation' stream not found. ✘"
                LOGGER.error(error_msg)
                self.error_signal.emit(error_msg)
                return
            inlet = lsl.StreamInlet(streams[0])
            msg = f"Stream: {streams[0].name} ✔"
            LOGGER.info(msg)
            self.found_signal.emit(msg)

            self.running = True
            self.id = threading.get_native_id()
            
            while self.running:
                sample, _ = inlet.pull_sample(timeout=0.1)
                if sample is None or len(sample) < 22:
                    continue
                current_time = sample[-2]
                screen_width, screen_height = sample[-4], sample[-3]
                
                gaze_x = int((sample[7] if sample[7] else sample[16]) * screen_width)
                gaze_y = int((sample[8] if sample[8] else sample[17]) * screen_height)
                
                self.buffer.append((current_time, gaze_x, gaze_y))

                fixation_time = sample[5] if sample[5] else sample[14]
                is_fixation = sample[3] or sample[12]
                if is_fixation and fixation_time:
                    if self.current_fixation is None:
                        self.current_fixation = {'x': gaze_x, 'y': gaze_y, 'count': 1, 'timestamp': fixation_time}
                    else:
                        self.current_fixation['x'] = gaze_x
                        self.current_fixation['y'] = gaze_y
                        self.current_fixation['count'] += 1
                elif self.current_fixation is not None:
                    self.fixation_buffer.append((
                        self.current_fixation['x'],
                        self.current_fixation['y'],
                        self.current_fixation['count'],
                        self.current_fixation['timestamp']
                    ))
                    self.current_fixation = None
                
            self.tracker_thread.stop()
            inlet.close_stream()
                        
        except Exception as e:
            LOGGER.error(f"Stream error: {str(e)}")
            self.error_signal.emit(f"Stream error: {str(e)}")

    def get_data(self, fixation=False):
        if fixation:
            return np.array(list(self.fixation_buffer), 
                            dtype=[('x', int), ('y', int), ('count', int), ('timestamp', float)])
        else:
            return np.array(list(self.buffer), 
                            dtype=[('timestamp', float), ('x', int), ('y', int)])
    
    def stop(self):
        self.running = False
        self.id = None
        self.buffer.clear()
        self.fixation_buffer.clear()
        self.current_fixation = None
        self.wait()
        self.quit()
        LOGGER.info("Stream thread stopped!")